/*
 * $Id: md.c,v 1.112.2.14 2008/02/29 07:02:45 spoel Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.3.3
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2008, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * Groningen Machine for Chemical Simulation
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gmx_fatal.h"
#include "typedefs.h"
#include "smalloc.h"
#include "sysstuff.h"
#include "vec.h"
#include "statutil.h"
#include "vcm.h"
#include "mdebin.h"
#include "nrnb.h"
#include "calcmu.h"
#include "index.h"
#include "vsite.h"
#include "update.h"
#include "ns.h"
#include "trnio.h"
#include "mdrun.h"
#include "confio.h"
#include "network.h"
#include "pull.h"
#include "xvgr.h"
#include "physics.h"
#include "names.h"
#include "xmdrun.h"
#include "disre.h"
#include "orires.h"
#include "dihre.h"
#include "pppm.h"
#include "pme.h"
#include "mdatoms.h"
#include "rmpbc.h"
#include "repl_ex.h"
#include "qmmm.h"
#include "compute_io.h"

#ifdef USE_MPE
#include "mpe.h"
#include "mpelogging.h"
#endif

/*>>> MFFA >>> */
typedef struct {
  real distance;
  real fric;
} FRIC;
/*<<< MFFA <<< */

volatile bool bGotTermSignal = FALSE, bGotUsr1Signal = FALSE;

static RETSIGTYPE signal_handler(int n)
{
  switch (n) {
  case SIGTERM:
    bGotTermSignal = TRUE;
    break;
  case SIGUSR1:
    bGotUsr1Signal = TRUE;
    break;
  }
}

/*>>> MFFA >>> */
void mpi_bound_force(PRES *part_Fb, real *Pb_sphere, rvec *Pb_plane, 
		int PNUMB, t_commrec *cr)
{
  int i, MESSAGE;
  real *buff;
#define SIZE 4
#define SIZE2 8
  
  MESSAGE = SIZE2 + (PNUMB*SIZE);
  snew(buff, MESSAGE);
  for(i=0;i<PNUMB;i++) {
	  buff[i*SIZE    ] = part_Fb[i].TFb_in;
	  buff[i*SIZE + 1] = part_Fb[i].TFb_out; 
	  buff[i*SIZE + 2] = part_Fb[i].TPb_in;
	  buff[i*SIZE + 3] = part_Fb[i].TPb_out;
  }
  
  /*Jump*/
  i = i * SIZE;
  buff[i    ] = Pb_sphere[0];
  buff[i + 1] = Pb_sphere[1];
  buff[i + 2] = Pb_plane[0][0];
  buff[i + 3] = Pb_plane[0][1];
  buff[i + 4] = Pb_plane[0][2];
  buff[i + 5] = Pb_plane[1][0];
  buff[i + 6] = Pb_plane[1][1];
  buff[i + 7] = Pb_plane[1][2];
  
  /*send and recieve message*/
  gmx_sumf(MESSAGE,buff,cr);
  
  for (i=0; i<PNUMB; i++) {
	  part_Fb[i].TFb_in  = buff[i*SIZE    ];
	  part_Fb[i].TFb_out = buff[i*SIZE + 1];
	  part_Fb[i].TPb_in  = buff[i*SIZE + 2];
	  part_Fb[i].TPb_out = buff[i*SIZE + 3];
  }
  
  /*jump*/
  i = i * SIZE;
  Pb_sphere[0]   = buff[i    ];
  Pb_sphere[1]   = buff[i + 1];
  Pb_plane[0][0] = buff[i + 2];
  Pb_plane[0][1] = buff[i + 3];
  Pb_plane[0][2] = buff[i + 4];
  Pb_plane[1][0] = buff[i + 5];
  Pb_plane[1][1] = buff[i + 6];
  Pb_plane[1][2] = buff[i + 7];
}

void correct_shell(bound_cuts *ref, bound_cuts *targ, real *shift_in, 
		real *shift_out) 
{
  /* hack for properly shifting thin shells used for restraining SHELL 
   * particles */
  real width;
  real middle;
  
  width = ref->pos_out - ref->pos_in;
  middle = ref->pos_in + 0.5*width;
  
  if(ref->pos_out < middle)
	  *shift_out = *shift_in;
  
  if(ref->pos_in > middle)
	  *shift_in = *shift_out;
}

void sdb_pc(int start, int nr_atoms, real *rout, real *rin, real refp, 
		real Pt, double dt, real tau, real *shift_out, real *shift_in, 
		real comp, rvec x[], rvec center)
{
  real dP,rout_old,rin_old;
  rvec mu;
  rvec dcent,old_cent,new_cent;
  int d,n,g;
  
  rout_old = *rout;
  rin_old = *rin;
  
  /* Pressure is now in bar, everywhere. */
  for (d=0; d<DIM; d++)
	  mu[d] = 1.0 - (comp*dt*(refp - Pt))/(DIM*tau);
  
  if (mu[XX]<0.99 || mu[XX]>1.01) 
	  fprintf(stderr,"Warning, scaling more than one pr.!!: mu %f\n", mu[XX]);
  
  /*calculate center shift*/
  new_cent[XX] = mu[XX]*center[XX];
  new_cent[YY] = mu[YY]*center[YY];
  new_cent[ZZ] = mu[ZZ]*center[ZZ];
  
  rvec_sub(center, new_cent, dcent);
  for (n=start; n<start+nr_atoms; n++) {
	  x[n][XX] = mu[XX]*x[n][XX] + dcent[XX];
	  x[n][YY] = mu[YY]*x[n][YY] + dcent[YY];
	  x[n][ZZ] = mu[ZZ]*x[n][ZZ] + dcent[ZZ];
  }
  
  *rout = mu[XX]*rout_old;
  *rin  = mu[XX]*rin_old;
  *shift_out = (*rout) - rout_old;
  *shift_in  = (*rin) - rin_old;
}

real calc_volume(MD_SD *SDB, tensor box, real *rad_in, real *rad_out)
{
  /* Note: It's not very efficient here to cast the whole struct MD_SD, 
   * better split it up in parts. however the effect on speed is very 
   * small */
  /* Note some set ups are not rock proof for every combination of potential 
   * type and/or index file !!! */
  real volume=0,r;
  int NUMB,i,axis;
  rvec new_box;
#define vsphere(r)((float)((4.0/3.0)*3.1415927*r*r*r))
#define vcilind(r,h)((float)(3.1415927*r*r*h))
  
  /* look at first sdb file only */
  if (SDB->bSDB == TRUE) {
	  NUMB=0;
	  axis=SDB->SDB[NUMB].AXIS;
	  if (SDB->SDB[NUMB].cutoffs.pos_in !=0)
		  *rad_in = SDB->SDB[NUMB].cutoffs.pos_in 
			  - SDB->SDB[NUMB].cutoffs.CUTOFF;
	  else
		  *rad_in = 0;
	  
	  if (SDB->SDB[NUMB].cutoffs.pos_out !=0)
		  *rad_out = SDB->SDB[NUMB].cutoffs.pos_out 
			  + SDB->SDB[NUMB].cutoffs.CUTOFF;
	  else
		  *rad_out=0;
	  
	  if (SDB->SDB[NUMB].TYPE == efSPHERE) {
		  if (*rad_out == 0)
			  volume = det(box) - vsphere(*rad_in);
		  /* difficult to say but I guess we have to treath it like a 
		   * big particle and put it's contributions in the virial*/
		  else
			  volume = vsphere(*rad_out)- vsphere(*rad_in);
	  } else if (SDB->SDB[NUMB].TYPE == efLIN) {
		  /* now a special case for test runs, a uniform planar boundary 
		   * condition without wild cards*/
		  if(!SDB->bNDX) {
			  for (i=0; i<DIM; i++)
				  new_box[i]=box[i][i];
			  new_box[axis] = *rad_out-*rad_in;
			  volume = new_box[0]*new_box[1]*new_box[2];
		  } else 
			  volume=det(box);
	  } else if (SDB->SDB[NUMB].TYPE == efCILIND) {
		  /* cilindric cavities */
		  if (!SDB->bNDX) {
			  if (SDB->SDB[NUMB].cutoffs.pos_out == 0)
				  volume = det(box)- vcilind(*rad_in,box[axis][axis]);
			  if (SDB->SDB[NUMB].cutoffs.pos_out != 0)
				  volume = vcilind(*rad_out,box[axis][axis]) 
					  - vcilind(*rad_in,box[axis][axis]);
		  } else
			  volume = det(box);
	  }
  } else
	  volume=det(box);
  
  return volume;
}

void split_string(char *string, char ***parts,int *numb)
{
  char *buf;
  int count, check=0, size=0, psize, i, next=0;
  
  /* number of potentials */
  snew(*parts,10);
  psize = 1;
  count = 0;
  size=(int)strlen(string);
  snew(buf, size+4); /* add 4 to size for backups _old */
  
  for (i=0; i<size; i++) {
	  if(string[i]=='#' && i!=0) {
		  next=0;
		  snew((*parts)[count],psize);
		  (*parts)[count] = strdup(buf);
		  /* clear buffer */
		  free(buf);
		  snew(buf,size);
		  psize = 0;
		  count++;
	  } else {
		  buf[next] = string[i];
		  next++;
		  psize++;
	  }
  }
  snew( (*parts)[count],psize);
  (*parts)[count] = strdup(buf);
  *numb = count+1;
}

/* interface for different potentials */
bool read_sdb_ndx(bool bNDB, char *filenm, int natoms, int ***SDB_IND, 
		int PNUMB)
{
  FILE *in;
  int i, j, loop=0, EXCLUDE=0;
  bool bWILD=FALSE, bEXCLUDE=FALSE;
  
  snew(*SDB_IND,natoms);

  if(bNDB) {
	  fprintf(stderr,"Reading index file.......\n");
	  in = ffopen(filenm,"r");
	  for (i=0; i<natoms; i++) {
		  fscanf(in,"%d\n",&loop);
		  /* actually this has to be the real size of loop for economical 
		   * reasons.....however this assymetry will you give a hard time 
		   * during communication */
		  /* Note: +1 slaves should be told this in mvdata.c */
		  snew((*SDB_IND)[i],PNUMB+1);
		  (*SDB_IND)[i][0]=loop;
		  bEXCLUDE = FALSE;
		  for (j=1; j<=loop; j++) {
			  fscanf(in,"%d\n", &(*SDB_IND)[i][j]);
			  /* check for wildcards */
			  if ((*SDB_IND)[i][j] >= PNUMB)
				  bWILD = TRUE;
			  if ((*SDB_IND)[i][j] < PNUMB)
				  bEXCLUDE = TRUE;
		  }
		  if (bEXCLUDE)
			  EXCLUDE++;
	  }
	  ffclose(in);
  } else {
	  /* default each potential works on each atom */
	  for (i=0; i<natoms; i++) {  
		  snew((*SDB_IND)[i],PNUMB+1);
		  (*SDB_IND)[i][0] = PNUMB;
		  for (j=1; j<=PNUMB; j++)
			  (*SDB_IND)[i][1]= j-1;
	  }
  }
  /* the first boundary potential is dominating the PC coupling, even if 
   * exclusion groups are used............ when acting on each atom the 
   * first potential gives a hard geometrical contrain on the volume */
  if(EXCLUDE == natoms)
	  bWILD == FALSE;
  /* Now bWILD is always false when a boundary potential works one every 
   * atom in the system !!!! */
  return bWILD;
}

void dump_sdb_settings(BOUNDS SDB,char *filenm, bool bQUENCH,real time)
{
  FILE *out;
  int i;
  
  out = fopen(filenm,"w");
  if (SDB.TYPE == efSPHERE) {
	  if(bQUENCH == FALSE)
		  fprintf(out,"SPHERE\n");
	  else 
		  fprintf(out,"QUENCH\n");
  } else if(SDB.TYPE == efLIN) {
	  fprintf(out,"LINEAR\n");
	  fprintf(out,"%c\n",88+SDB.AXIS);
  } else if(SDB.TYPE == efCILIND) {
	  fprintf(out,"CILIND\n");
	  fprintf(out,"%c\n",88+SDB.AXIS);
  }
  if (SDB.cutoffs.pos_in != 0)
	  fprintf(out,"%f\n", (real)(SDB.cutoffs.pos_in  - SDB.cutoffs.CUTOFF));
  else
	  fprintf(out,"%f\n", 0.0);
  
  if (SDB.cutoffs.pos_out != 0)
	  fprintf(out,"%f\n", (real)(SDB.cutoffs.pos_out + SDB.cutoffs.CUTOFF));
  else
	  fprintf(out,"%f\n", 0.0);
  
  fprintf(out,"%f\n", SDB.cutoffs.CUTOFF);
  fprintf(out,"%f\n",SDB.cutoffs.GRID);
  fprintf(out,"%f\n",SDB.cutoffs.a_in);
  fprintf(out,"%f\n",SDB.cutoffs.a_out);
  fprintf(out,"%f\n",SDB.cutoffs.center[XX]);
  fprintf(out,"%f\n",SDB.cutoffs.center[YY]);
  fprintf(out,"%f\n",SDB.cutoffs.center[ZZ]);
  
  if (SDB.TYPE == efSPHERE)
	  fprintf(out,"%c type\n%c inner cutoff\n%c outer cutoff\n%c cutoff\n%c gridsize\n%c a in\n%c a out\n%c center X Y Z\n",37,37,37,37,37,37,37,37);
  else
	  fprintf(out,"%c type\n%c Axis\n%c inner cutoff\n%c outer cutoff\n%c cutoff\n%c gridsize\n%c a in (=> 0)\n%c a out (=>0)\n%c center X Y Z\n",37,37,37,37,37,37,37,37,37);
  
  fprintf(out,"%c version 3.3, 10 9 2006\n",37);
  fprintf(out,"%c t = %8.3f (ns)\n",37,time);
  fprintf(out,"#\n");
  for(i=0;i<SDB.cutoffs.index_nr_in;i++)
	  fprintf(out,"%f\n",SDB.F_bounds[i]);
  fprintf(out,"#\n");
  for(;i<SDB.cutoffs.index_nr_out;i++)
	  fprintf(out,"%f\n",SDB.F_bounds[i]);
  fprintf(out,"#\n");
  fclose(out);
}

void read_sdb(char *gfile, BOUNDS *SDB, int *TYPE, int *bQUENCH)
{
  int i=0, count, count2, check;
  real *buffer, number;
  FILE *in;
  fpos_t pos0, pos1, pos2;
  char c='a', type[10], buf[STRLEN], axtitle[STRLEN], VERS[STRLEN];
  bool loop=TRUE, loop2=TRUE, First1=TRUE, First2=TRUE, bCHECK=TRUE;
  real cutoff, radin, radout;
  
  if ((in=fopen(gfile,"r")) == NULL)
	  gmx_fatal(FARGS,"No .sdb file found!!!!!!!!\n");
  
  fprintf(stderr,"Getting data form .sdb file\n");
  fscanf(in,"%s",type); 
  *bQUENCH = 0;
  if (strncmp(type,"SPHERE",6)==0)
	  *TYPE = efSPHERE;
  else if (strncmp(type,"LINEAR",6)==0)
	  *TYPE = efLIN;
  else if (strncmp(type,"QUENCH",6)==0) {
	  *TYPE = efSPHERE;
	  *bQUENCH = 1;
  } else if(strncmp(type,"CILIND",6)==0)
	  *TYPE = efCILIND;
  else
	  bCHECK =FALSE;
  
  if (!bCHECK)
	  gmx_fatal(FARGS,"Type is not correct specified in %s file\n",gfile);
  else
	  fprintf(stderr,"Selected potential type: %s\n",type);
  
  if(*TYPE == efLIN || *TYPE == efCILIND) {
	  fscanf(in,"%s",axtitle);
	  SDB->AXIS = toupper(axtitle[0]) - 'X';
	  if(SDB->AXIS < 0 || SDB->AXIS > 2)
		  gmx_fatal(FARGS,"Axis is not correct specified in %s file\n",gfile);
	  fprintf(stderr,"Potential along %c-axis\n",axtitle[0]);
  }
  fscanf(in,"%f",&radin);
  fscanf(in,"%f",&radout);
  fscanf(in,"%f",&cutoff);
  
  SDB->cutoffs.CUTOFF = cutoff;
  if (radin!=0)
	  SDB->cutoffs.pos_in=(real)(radin+cutoff);
  else
	  SDB->cutoffs.pos_in=0;
  fprintf(stderr,"pos_in set to from center: %f\n", SDB->cutoffs.pos_in);
  if(radout!=0)
	  SDB->cutoffs.pos_out=(real)(radout-cutoff);
  else
	  SDB->cutoffs.pos_out=0;
  fprintf(stderr,"pos_out set to from center: %f\n", SDB->cutoffs.pos_out);
  
  fscanf(in,"%f",&SDB->cutoffs.GRID);
  fprintf(stderr,"GRID-size set to: %f\n", SDB->cutoffs.GRID);
  fscanf(in,"%f",&SDB->cutoffs.a_in);
  fprintf(stderr,"a_in set to: %f\n", SDB->cutoffs.a_in);
  fscanf(in,"%f",&SDB->cutoffs.a_out);
  fprintf(stderr,"a_out set to: %f\n", SDB->cutoffs.a_out);
  fscanf(in,"%f",&SDB->cutoffs.center[XX]);
  fscanf(in,"%f",&SDB->cutoffs.center[YY]);
  fscanf(in,"%f",&SDB->cutoffs.center[ZZ]);
  fprintf(stderr,"center set to: %f %f %f\n", SDB->cutoffs.center[XX], 
		  SDB->cutoffs.center[YY], SDB->cutoffs.center[ZZ]);
  /* check version */
  /* newer version shorter length */
  fscanf(in,"%s",VERS);
  if (strncmp(VERS,"%",1)!=0 && strncmp(VERS,"#",1) != 0)
	  gmx_fatal(FARGS, "Format not correct! Seems you are using an old .sdb version\n");
  
  /* make sure a_in a_out are not negative */
  SDB->cutoffs.a_in =fabs(SDB->cutoffs.a_in);
  SDB->cutoffs.a_out=fabs(SDB->cutoffs.a_out);  
  
  fprintf(stderr,"Reading coordinates for force......\n");
  check=0;
  count=-1;
  count2=-1;
  rewind(in);
  while ((c = getc(in))!=EOF && check != 3) {
	  if (c=='#')
		  check++;
	  if (check==1) {
		  if (First1)
			  fgetpos(in,&pos1);
		  if (c=='\n')
			  count++;
		  First1=FALSE;
	  }
	  if (check==2) {
		  if (First2) 
			  fgetpos(in,&pos2);
		  if (c=='\n')
			  count2++;
		  First2=FALSE;
	  }
  }
  
  if (check!=3)
	  gmx_fatal(FARGS,"Less than 3 # characters in .sdb file");
  
  SDB->cutoffs.index_nr_in = count;
  SDB->cutoffs.index_nr_out = count2 + count;
  fprintf(stderr,"Read %d coordinates for in and %d for out.\n",count,count2);
  
  snew(SDB->F_bounds,(count+count2));
  fsetpos(in,&pos1);
  for (i=0;i<count;i++) {
	  fscanf(in,"%f\n",&SDB->F_bounds[i]);
	  fprintf(stderr, "inner force: %f  %f\n", 
			  (float)(SDB->cutoffs.pos_in 
				  - (SDB->cutoffs.GRID * (SDB->cutoffs.index_nr_in - i))), 
			  SDB->F_bounds[i]);
  }
  fsetpos(in,&pos2);
  for (i=0;i<count2;i++) {
	  fscanf(in,"%f\n",&SDB->F_bounds[i+count]);
	  fprintf(stderr, "outer force: %f  %f\n", 
			  (float)((SDB->cutoffs.GRID*i) + SDB->cutoffs.pos_out), 
			  SDB->F_bounds[i+count]);
  }
  fclose(in);
}

void do_sdb_force(int start, int homenr, rvec f[], rvec x[], MD_SD *SDB, 
		matrix box, matrix force_vir)
{
  int j, m, n, d, k, index, NUMB=0, i, axis;
  real distance=0, fit, hbox=0, shift=0, boxz=0;
  rvec distvec, f_old, box_vec;
  bool bWILDCARD=TRUE, bEMPTY;
  real RIN=0, ROUT=0, bforce=0, pot_dist=0, LAST_IN=0, LAST_OUT=0;
  matrix bound_vir;
  static bool bFIRST = FALSE;
  
  /* clear arrays for pressures and forces */
  clear_mat(bound_vir);
  for (k=0; k<SDB->PNUMB; k++) { 
	  SDB->part_Fb[k].TFb_in  = 0;
	  SDB->part_Fb[k].TFb_out = 0;
	  SDB->part_Fb[k].TPb_in  = 0;
	  SDB->part_Fb[k].TPb_out = 0;
  }
  
  clear_rvec(SDB->Pb_plane[0]);
  clear_rvec(SDB->Pb_plane[1]);
  SDB->Pb_sphere[0] = 0;
  SDB->Pb_sphere[1] = 0;
  
  /* loop over n atoms */
  for (n=start; n<start+homenr; n++) { 
	  /* store old force */
	  for (d=0; d<DIM; d++)
		  f_old[d] = f[n][d];
	  for (k=1; k<=SDB->SDB_IND[n][0]; k++) { 
		  clear_rvec(distvec);
		  bWILDCARD = TRUE;
		  NUMB = SDB->SDB_IND[n][k];
		  if (NUMB >= SDB->PNUMB)
			  bWILDCARD = FALSE;
		  /* If FALSE; don't touch me............. */
		  /* reset bforce */
		  bforce = 0;
		  if (bWILDCARD) {
			  bEMPTY = FALSE;
			  /* hack to prevent empty arrays */
			  if (SDB->SDB[NUMB].cutoffs.index_nr_out != 0)
				  LAST_OUT = SDB->SDB[NUMB].F_bounds[SDB->SDB[NUMB].cutoffs.index_nr_out - 1];
			  else {
				  LAST_OUT = 0; 
				  bEMPTY = TRUE;
			  }
			  if (SDB->SDB[NUMB].cutoffs.index_nr_in != 0)
				  LAST_IN = SDB->SDB[NUMB].F_bounds[0];
			  else
				  LAST_IN = 0;
			  if (SDB->SDB[NUMB].TYPE == efSPHERE) {
				  /* spheres */
				  for (d=0; d<DIM; d++)
					  distvec[d] = x[n][d] - SDB->SDB[NUMB].cutoffs.center[d];
				  distance = sqrt(iprod(distvec,distvec));
				  /* unit normal vec */
				  for (d=0; d<DIM; d++)
					  distvec[d] = distvec[d]/distance;
			  } else if (SDB->SDB[NUMB].TYPE == efLIN) {
				  /* planes */
				  axis = SDB->SDB[NUMB].AXIS;
				  distance = x[n][axis];
				  /* unit normal vec */
				  clear_rvec(distvec);
				  distvec[axis] = 1.0;
			  } else if (SDB->SDB[NUMB].TYPE == efCILIND) {
				  /* Cylinder */
				  for (d=0; d<DIM; d++)
					  distvec[d] = x[n][d] - SDB->SDB[NUMB].cutoffs.center[d];
				  axis = SDB->SDB[NUMB].AXIS;
				  /* cilindric axis */
				  distvec[axis] = 0;
				  distance = sqrt(iprod(distvec,distvec));
				  /* unit normal vec */
				  for (d=0; d<DIM; d++)
					  distvec[d] = distvec[d]/distance;
			  } else
				  gmx_fatal(FARGS,"Type error in one of the .sdb files!\n");
			  if (SDB->SDB[NUMB].cutoffs.pos_out != 0)
				  /* check outher boundary is switched on anyway */
				  if (distance > SDB->SDB[NUMB].cutoffs.pos_out) {
					  index = ((int)(
								  (distance - SDB->SDB[NUMB].cutoffs.pos_out) 
								  / SDB->SDB[NUMB].cutoffs.GRID)) 
						  + SDB->SDB[NUMB].cutoffs.index_nr_in;
					  if (index < SDB->SDB[NUMB].cutoffs.index_nr_out 
							  && !bEMPTY) {
						  bforce = SDB->SDB[NUMB].F_bounds[index];
						  SDB->part_Fb[NUMB].TFb_out += bforce; 
						  for (d=0; d<DIM; d++)
							  f[n][d] += distvec[d]*bforce;
					  } else {
						  /* a_in > 0 pot_dist > 0 */
						  pot_dist = 
							  (distance - SDB->SDB[NUMB].cutoffs.pos_out) 
							  - ((SDB->SDB[NUMB].cutoffs.index_nr_out 
										  - SDB->SDB[NUMB].cutoffs.index_nr_in) 
									  * SDB->SDB[NUMB].cutoffs.GRID);
						  fit = LAST_OUT - (SDB->SDB[NUMB].cutoffs.a_out 
								  * pot_dist);
						  bforce = fit;
						  SDB->part_Fb[NUMB].TFb_out += bforce;
						  for (d=0; d<DIM; d++)
							  f[n][d] += distvec[d]*bforce;
					  }
					  for (j=0; j < DIM; j++)
						  for (m=0; m < DIM; m++)
							  bound_vir[j][m] += bforce * distvec[j] 
								  * distvec[m] 
								  * (SDB->SDB[NUMB].cutoffs.pos_out 
										  + SDB->SDB[NUMB].cutoffs.CUTOFF 
										  - distance);
				  }
			  if (distance < SDB->SDB[NUMB].cutoffs.pos_in) {
				  index = (int)(((distance - SDB->SDB[NUMB].cutoffs.pos_in) 
							  / SDB->SDB[NUMB].cutoffs.GRID) 
						  + SDB->SDB[NUMB].cutoffs.index_nr_in);
				  if (index<0) {
					  /* a_in > 0 and pot_dist > 0 */ 
					  pot_dist = (SDB->SDB[NUMB].cutoffs.pos_in - distance) 
						  - (SDB->SDB[NUMB].cutoffs.index_nr_in 
								  * SDB->SDB[NUMB].cutoffs.GRID);
					  fit = LAST_IN - (SDB->SDB[NUMB].cutoffs.a_in * pot_dist);
					  bforce = fit;
					  SDB->part_Fb[NUMB].TFb_in  += bforce;
					  for (d=0; d<DIM; d++)
						  f[n][d] += distvec[d] * bforce * -1.0;
				  } else if (index< SDB->SDB[NUMB].cutoffs.index_nr_in) {
					  bforce = SDB->SDB[NUMB].F_bounds[index];
					  SDB->part_Fb[NUMB].TFb_in += bforce;
					  for (d=0; d<DIM; d++)
						  f[n][d] += distvec[d]* bforce * -1.0;
				  }
				  for (j=0; j < DIM; j++)
					  for (m=0; m < DIM; m++)
						  bound_vir[j][m] += bforce * distvec[j] 
							  * distvec[m] * (distance 
									  - SDB->SDB[NUMB].cutoffs.pos_in 
									  + SDB->SDB[NUMB].cutoffs.CUTOFF);
			  }
		  }
	  }
	  /**********uniform***********/
  }
  /* add boundary contributions to virial */
  m_add(bound_vir,force_vir,force_vir);
  /* Boundary pressure, AREA, etc */
  ROUT=0;
  RIN=0;
#define asphere(r)((float)(4.0*3.1415927*r*r))
#define laplace(r,s)((float)((s*2)/r))
#define acilind(r,h)((float)(3.1415927*2.0*r*h))
  
  for (k=0; k< SDB->PNUMB; k++) {
	  if (SDB->SDB[k].cutoffs.pos_in != 0)
		  RIN = SDB->SDB[k].cutoffs.pos_in - SDB->SDB[k].cutoffs.CUTOFF;
	  else
		  RIN  = 0;
	  if (SDB->SDB[k].cutoffs.pos_out != 0)
		  ROUT = SDB->SDB[k].cutoffs.pos_out + SDB->SDB[k].cutoffs.CUTOFF;
	  else
		  ROUT = 0;
	  
	  if (SDB->SDB[k].TYPE == efLIN) {
		  for (d=0; d<DIM; d++)
			  box_vec[d] = box[d][d];
		  box_vec[SDB->SDB[k].AXIS] = 1.0;
		  /* rectangular only */
		  SDB->part_Fb[k].AREA_in = SDB->part_Fb[k].AREA_out 
			  = box_vec[0]*box_vec[1]*box_vec[2];
	  } else if (SDB->SDB[k].TYPE == efSPHERE) {
		  SDB->part_Fb[k].AREA_in = asphere(RIN);
		  SDB->part_Fb[k].AREA_out = asphere(ROUT);
	  } else if (SDB->SDB[k].TYPE == efCILIND) {
		  boxz = box_vec[SDB->SDB[k].AXIS];
		  SDB->part_Fb[k].AREA_in = acilind(RIN, boxz);
		  SDB->part_Fb[k].AREA_out = acilind(ROUT,boxz);
	  }
	  if (RIN != 0)
		  SDB->part_Fb[k].TPb_in = -1.0 * PRESFAC* (SDB->part_Fb[k].TFb_in 
				  / SDB->part_Fb[k].AREA_in);
	  else
		  SDB->part_Fb[k].TPb_in = 0;
	  
	  if (SDB->SDB[k].TYPE == efSPHERE)
		  SDB->Pb_sphere[0] += SDB->part_Fb[k].TPb_in;
	  else if (SDB->SDB[k].TYPE == efLIN)
		  SDB->Pb_plane[0][SDB->SDB[k].AXIS] += SDB->part_Fb[k].TPb_in;
	  
	  /* total uniform */
	  if (ROUT != 0)
		  SDB->part_Fb[k].TPb_out = -1.0 * PRESFAC 
			  * (SDB->part_Fb[k].TFb_out / SDB->part_Fb[k].AREA_out);
	  else
		  SDB->part_Fb[k].TPb_out = 0;
	  /* total uniform */
	  if (SDB->SDB[k].TYPE == efSPHERE)
		  SDB->Pb_sphere[1] += SDB->part_Fb[k].TPb_out;
	  else if (SDB->SDB[k].TYPE == efLIN)
		  SDB->Pb_plane[1][SDB->SDB[k].AXIS] += SDB->part_Fb[k].TPb_out;
  }
  bFIRST = TRUE;
}
/*<<< MFFA <<<*/

void mdrunner(t_commrec *cr,t_commrec *mcr,int nfile,t_filenm fnm[],
	      bool bVerbose,bool bCompact,
	      int nDlb,int nstepout,
	      t_edsamyn *edyn,int repl_ex_nst,int repl_ex_seed,
	      unsigned long Flags)
{
  double     nodetime=0,realtime;
  t_inputrec *inputrec;
  t_state    *state;
  rvec       *buf,*f,*vold,*vt;
  real       tmpr1,tmpr2;
  real       *ener;
  t_nrnb     *nrnb;
  t_nsborder *nsb;
  t_topology *top;
  t_groups   *grps;
  t_graph    *graph;
  t_mdatoms  *mdatoms;
  t_forcerec *fr;
  t_fcdata   *fcd;
  time_t     start_t=0;
  bool       bVsites,bParVsites;
  t_comm_vsites vsitecomm;
  int        i,m;
  char       *gro;
  /*>>> MFFA >>>*/
  /* spherical boundary stuff */
  char **SDB_FILE_PARTS=NULL;
  MD_SD *SDB=NULL;
  /*<<< MFFA <<<*/
  
  /* Initiate everything (snew sets to zero!) */
  snew(ener,F_NRE);
  snew(fcd,1);
  snew(nsb,1);
  snew(top,1);
  snew(grps,1);
  snew(inputrec,1);
  snew(state,1);
  snew(nrnb,cr->nnodes);
  /*>>> MFFA >>>*/
  snew(SDB,1);
  SDB->bSDB = FALSE;
  SDB->bPC  = FALSE;
  /*<<< MFFA <<<*/
  
  if (bVerbose && MASTER(cr)) 
    fprintf(stderr,"Getting Loaded...\n");

  if (PAR(cr)) {
    /* The master thread on the master node reads from disk, then passes everything
     * around the ring, and finally frees the stuff
     */
    if (MASTER(cr)) 
      distribute_parts(cr->left,cr->right,cr->nodeid,cr->nnodes,inputrec,
		       ftp2fn(efTPX,nfile,fnm),nDlb);
    
    /* Every node (including the master) reads the data from the ring */
    init_parts(stdlog,cr,
	       inputrec,top,state,&mdatoms,nsb,
	       MASTER(cr) ? LIST_SCALARS | LIST_INPUTREC : 0,
	       &bParVsites,&vsitecomm);
  } else {
    /* Read it up... */
    init_single(stdlog,inputrec,ftp2fn(efTPX,nfile,fnm),top,state,&mdatoms,nsb);
    bParVsites=FALSE;
  }
  if (inputrec->eI == eiSD) {
    /* Is not read from TPR yet, so we allocate space here */
    snew(state->sd_X,nsb->natoms);
  }
  snew(buf,nsb->natoms);
  snew(f,nsb->natoms);
  snew(vt,nsb->natoms);
  snew(vold,nsb->natoms);

  if (bVerbose && MASTER(cr))
    fprintf(stderr,"Loaded with Money\n\n");
  
  /* Index numbers for parallellism... */
  nsb->nodeid      = cr->nodeid;
  top->idef.nodeid = cr->nodeid;

  /* Group stuff (energies etc) */
  init_groups(stdlog,mdatoms,&(inputrec->opts),grps);
  /* Copy the cos acceleration to the groups struct */
  grps->cosacc.cos_accel = inputrec->cos_accel;
  
  /* Periodicity stuff */  
  if (inputrec->ePBC == epbcXYZ) {
    graph=mk_graph(&(top->idef),top->atoms.nr,FALSE,FALSE);
    if (debug)
      p_graph(debug,"Initial graph",graph);
  }
  else
    graph = NULL;
    
  /* Distance Restraints */
  init_disres(stdlog,top->idef.il[F_DISRES].nr,top->idef.il[F_DISRES].iatoms,
	      top->idef.iparams,inputrec,mcr,fcd);

  /* Orientation restraints */
  init_orires(stdlog,top->idef.il[F_ORIRES].nr,top->idef.il[F_ORIRES].iatoms,
	      top->idef.iparams,state->x,mdatoms,inputrec,mcr,
	      &(fcd->orires));

  /* Dihedral Restraints */
  init_dihres(stdlog,top->idef.il[F_DIHRES].nr,top->idef.il[F_DIHRES].iatoms,
	      top->idef.iparams,inputrec,fcd);

  /* check if there are vsites */
  bVsites=FALSE;
  for(i=0; (i<F_NRE) && !bVsites; i++)
    bVsites = ((interaction_function[i].flags & IF_VSITE) && 
		(top->idef.il[i].nr > 0));

  /* Initiate forcerecord */
  fr = mk_forcerec();
  init_forcerec(stdlog,fr,inputrec,top,cr,mdatoms,nsb,state->box,FALSE,
		opt2fn("-table",nfile,fnm),opt2fn("-tablep",nfile,fnm),FALSE);
  fr->bSepDVDL = ((Flags & MD_SEPDVDL) == MD_SEPDVDL);

  /* Initialize QM-MM */
  if(fr->bQMMM){
    init_QMMMrec(cr,mdatoms,state->box,top,inputrec,fr);
  }
   
  /* Initiate PPPM if necessary */
  if (fr->eeltype == eelPPPM)
    init_pppm(stdlog,cr,nsb,FALSE,TRUE,state->box,getenv("GMXGHAT"),inputrec);
  if ((fr->eeltype == eelPME) || (fr->eeltype == eelPMEUSER))
    (void) init_pme(stdlog,cr,inputrec->nkx,inputrec->nky,inputrec->nkz,
		    inputrec->pme_order,
		    /*HOMENR(nsb),*/nsb->natoms,
		    mdatoms->bChargePerturbed,
		    inputrec->bOptFFT,inputrec->ewald_geometry);
  
  /*>>> MFFA >>>*/
  /* initialise SDB routines */
  if (MASTER(cr)) {
	  if (opt2bSet("-sb",nfile,fnm)) {
		  split_string(opt2fn("-sb",nfile,fnm),&SDB_FILE_PARTS,&SDB->PNUMB);
		  for (i=0; i<SDB->PNUMB; i++)
			  fprintf(stderr,"Found: %s\n ",SDB_FILE_PARTS[i]);
		  snew(SDB->SDB,SDB->PNUMB);
		  int QUENCH=0, COOL=0;
		  if(SDB->PNUMB>=1) {
			  for(i=0; i<SDB->PNUMB;) {
				  read_sdb(SDB_FILE_PARTS[i], &(SDB->SDB[i]), 
						  &(SDB->SDB[i].TYPE), &QUENCH);
				  COOL += QUENCH;
				  i++;
			  }
		  } else
			  gmx_fatal(FARGS,"No .sdb file found!!!!!!!!\n");
		  if (COOL >= 1) {
			  SDB->bQUENCH = TRUE;
			  fprintf(stderr,"Quench routine set to active!\n Warning, this will be subjected to all spherical shells within in the simulation!\n");
		  } else if (COOL == 0)
			  SDB->bQUENCH = FALSE;
		  
		  SDB->bSDB=TRUE;
		  SDB->bNDX=opt2bSet("-ndb",nfile,fnm);

		  /* note that SDB->bNDX is used for wildcard check after returning */
		  SDB->bNDX = read_sdb_ndx(SDB->bNDX, opt2fn("-ndb",nfile,fnm), 
				  nsb->natoms, &(SDB->SDB_IND), SDB->PNUMB);
		  if (SDB->bNDX)
			  fprintf(stderr,"Found boundary exclusion group in index file for first potential!\n");
		  
		  SDB->size_ndx = nsb->natoms;
	  }
  }
  /* mpi communications */
  if (PAR(cr))
	  distribute_bound(cr->left,cr->right,SDB,cr);
  if (SDB->bSDB) {
	  /* only for union shell systems in this implementation */
	  snew(SDB->part_Fb,SDB->PNUMB);
	  if (SDB->SDB[0].cutoffs.pos_out != 0 && SDB->SDB[0].TYPE == efSPHERE) {
		  if (fr->ePBC == epbcNONE)
			  if (inputrec->ref_p[XX][XX] != 0) {
				  if (MASTER(cr))
					  fprintf(stderr,"SDB-pressure coupling (on Virial) switched on!!!\n");
				  SDB->bPC=TRUE;
			  }
		  if (inputrec->epc != epcNO) {
			  /* NVT with pressure : although its thermodynamicle bullshit */
			  if (inputrec->ref_p[XX][XX] == 0)
				  inputrec->compress[XX][XX] = 0;
			  if (MASTER(cr))
				  fprintf(stderr,"SDB-pressure coupling (on virial) switched on!!!\n");
			  SDB->bPC=TRUE;
		  }
	  }
	  /* put normal routine out for sphere */
	  if (SDB->bPC)
		  inputrec->epc=epcNO;
  }
  /*<<< MFFA <<<*/

  /* Make molecules whole at start of run */
  if (fr->ePBC != epbcNONE) {
	  do_pbc_first(stdlog,state->box,fr,graph,state->x);
  }

  /* Now do whatever the user wants us to do (how flexible...) */
  switch (inputrec->eI) {
  case eiMD:
  case eiSD:
  case eiBD:
    start_t=do_md(stdlog,cr,mcr,nfile,fnm,
		  bVerbose,bCompact,bVsites,
		  bParVsites ? &vsitecomm : NULL,
		  nstepout,inputrec,grps,top,ener,fcd,state,vold,vt,f,buf,
		  mdatoms,nsb,nrnb,graph,edyn,fr,
		  repl_ex_nst,repl_ex_seed,
		  Flags, SDB); /*<<< MFFA >>>*/
    break;
  case eiCG:
    start_t=do_cg(stdlog,nfile,fnm,inputrec,top,grps,nsb,
		  state,f,buf,mdatoms,ener,fcd,
		  nrnb,bVerbose,bVsites,
		  bParVsites ? &vsitecomm : NULL,
		  cr,mcr,graph,fr);
    break;
  case eiLBFGS:
    start_t=do_lbfgs(stdlog,nfile,fnm,inputrec,top,grps,nsb,
		     state,f,buf,mdatoms,ener,fcd,
		     nrnb,bVerbose,bVsites,
		     bParVsites ? &vsitecomm : NULL,
		     cr,mcr,graph,fr);
    break;
  case eiSteep:
    start_t=do_steep(stdlog,nfile,fnm,inputrec,top,grps,nsb,
		     state,f,buf,mdatoms,ener,fcd,
		     nrnb,bVerbose,bVsites,
		     bParVsites ? &vsitecomm : NULL,
		     cr,mcr,graph,fr);
    break;
  case eiNM:
    start_t=do_nm(stdlog,cr,nfile,fnm,
		  bVerbose,bCompact,nstepout,inputrec,grps,
		  top,ener,fcd,state,vold,vt,f,buf,
		  mdatoms,nsb,nrnb,graph,edyn,fr);
    break;
  case eiTPI:
    start_t=do_tpi(stdlog,nfile,fnm,inputrec,top,grps,nsb,
		   state,f,buf,mdatoms,ener,fcd,
		   nrnb,bVerbose,
		   cr,mcr,graph,fr);
    break;
  default:
    gmx_fatal(FARGS,"Invalid integrator (%d)...\n",inputrec->eI);
  }
  
    /* Some timing stats */  
  if (MASTER(cr)) {
    realtime=difftime(time(NULL),start_t);
    if ((nodetime=node_time()) == 0)
      nodetime=realtime;
  }
  else 
    realtime=0;
    
  /* Convert back the atoms */
  md2atoms(mdatoms,&(top->atoms),TRUE);
  
  /* Finish up, write some stuff
   * if rerunMD, don't write last frame again 
   */
  finish_run(stdlog,cr,ftp2fn(efSTO,nfile,fnm),
	     nsb,top,inputrec,nrnb,nodetime,realtime,inputrec->nsteps,
	     EI_DYNAMICS(inputrec->eI));
  
  /* Does what it says */  
  print_date_and_time(stdlog,cr->nodeid,"Finished mdrun");
}

time_t do_md(FILE *log,t_commrec *cr,t_commrec *mcr,int nfile,t_filenm fnm[],
	     bool bVerbose,bool bCompact,
	     bool bVsites, t_comm_vsites *vsitecomm,
	     int stepout,t_inputrec *inputrec,t_groups *grps,t_topology *top,
	     real ener[],t_fcdata *fcd,
	     t_state *state,rvec vold[],rvec vt[],rvec f[],
	     rvec buf[],t_mdatoms *mdatoms,t_nsborder *nsb,t_nrnb nrnb[],
	     t_graph *graph,t_edsamyn *edyn,t_forcerec *fr,
	     int repl_ex_nst,int repl_ex_seed,
	     unsigned long Flags, MD_SD *pdsb) /*<<< MFFA >>>*/
{
  t_mdebin   *mdebin;
  int        fp_ene=0,fp_trn=0,step,step_rel;
  FILE       *fp_dgdl=NULL,*fp_field=NULL;
  time_t     start_t;
  real       t,t0,lam0;
  bool       bNS,bSimAnn,bStopCM,bRerunMD,bNotLastFrame=FALSE,
             bFirstStep,bLastStep,bNEMD,do_log,bRerunWarnNoV=TRUE,
	     bFullPBC,bForceUpdate=FALSE;
  tensor     force_vir,shake_vir,total_vir,pres,ekin;
  t_nrnb     mynrnb;
  char       *traj,*xtc_traj; /* normal and compressed trajectory filename */
  int        i,m,status;
  rvec       mu_tot;
  rvec       *xx,*vv,*ff;
  t_vcm      *vcm;
  t_trxframe rerun_fr;
  t_pull     pulldata; /* for pull code */
  gmx_repl_ex_t *repl_ex=NULL;
  /* A boolean (disguised as a real) to terminate mdrun */  
  real       terminate=0;
 
  /* XMDRUN stuff: shell, general coupling etc. */
  bool        bFFscan;
  int         nshell,nflexcon,nshell_flexcon_tot,count,nconverged=0;
  t_shell     *shells=NULL;
  real        timestep=0;
  double      tcount=0;
  bool        bShell_FlexCon,bIonize=FALSE,bGlas=FALSE;
  bool        bTCR=FALSE,bConverged=TRUE,bOK,bExchanged;
  real        temp0,mu_aver=0,dvdl;
  int         gnx,ii;
  atom_id     *grpindex;
  char        *grpname;
  t_coupl_rec *tcr=NULL;
  rvec        *xcopy=NULL,*vcopy=NULL, center, PARM;
  matrix      boxcopy,lastbox;
  /* End of XMDRUN stuff */
  /*>>> MFFA >>>*/
  /* some running parameters for SDB hacks */
  real P_out_tot=0, P_out_interval=0;
  real P_in_tot=0, P_in_interval=0;
  real P_vir_interval=0, P_vir_tot =0;
  int count1=0, count2=0, n, j;
  real shift_out, shift_in, pos_out_ref=0, pos_in_ref=0;
  real volume=0, rin, rout, *buff, AREA=0;
  bool bUPDSDB;
  
  snew(buff,2);
  bUPDSDB = (inputrec->compress[XX][XX] == 0 && pdsb->bSDB);
  /*<<< MFFA <<<*/

  /* Turn on signal handling */
  if (getenv("GMX_NO_TERM") == NULL)
    signal(SIGTERM,signal_handler);
  if (getenv("GMX_NO_USR1") == NULL)
    signal(SIGUSR1,signal_handler);

  /* Check for special mdrun options */
  bRerunMD = (Flags & MD_RERUN)  == MD_RERUN;
  bIonize  = (Flags & MD_IONIZE) == MD_IONIZE;
  bGlas    = (Flags & MD_GLAS)   == MD_GLAS;
  bFFscan  = (Flags & MD_FFSCAN) == MD_FFSCAN;
  
  /*>>> MFFA >>>*/
  /* piston parameters */
  PARM[0] = inputrec->userreal2;
  PARM[1] = inputrec->userreal3;
  /* this ia actually an int */
  PARM[2] = inputrec->userreal4;
  /* Make sure rotational momentum is removed !!!!!!!!!!!!*/
  if(pdsb->bSDB)
	  if(pdsb->SDB[0].TYPE == efSPHERE) {
		  inputrec->comm_mode = ecmANGULAR;
		  inputrec->nstcomm = 1;
		  if (MASTER(cr))
			  fprintf(stderr,"Simulating isolated cluster in vacuum. Angular momentum will be removed at every step !\n");
	  }
  /*<<< MFFA <<<*/

  /* Initial values */
  init_md(cr,inputrec,&t,&t0,&state->lambda,&lam0,&mynrnb,top,
	  nfile,fnm,&traj,&xtc_traj,&fp_ene,&fp_dgdl,&fp_field,&mdebin,grps,
	  force_vir,shake_vir,mdatoms,mu_tot,&bNEMD,&bSimAnn,&vcm,nsb);
  debug_gmx();

  {
    double io = compute_io(inputrec,&top->atoms,mdebin->ebin->nener,1);
    if (io > 2000) 
      fprintf(stderr,
	      "\nWARNING: This run will generate roughly %.0f Mb of data\n\n",
	      io);
  }
  
  /* init edsam, no effect if edyn->bEdsam==FALSE */
  init_edsam(stdlog,top,inputrec,mdatoms,START(nsb),HOMENR(nsb),cr,edyn);
    
  /* Check for full periodicity calculations */
  bFullPBC = (inputrec->ePBC == epbcFULL);  
  
  /* Check for polarizable models */
  shells     = init_shells(log,START(nsb),HOMENR(nsb),&top->idef,
			   mdatoms,&nshell);

  /* Check for flexible constraints */
  nflexcon = count_flexible_constraints(log,fr,&top->idef);

  nshell_flexcon_tot = nshell + nflexcon;
  if (PAR(cr))
    gmx_sumi(1,&nshell_flexcon_tot,cr);
  bShell_FlexCon = nshell_flexcon_tot > 0;
  
  gnx = top->blocks[ebMOLS].nr;
  snew(grpindex,gnx);
  for(i=0; (i<gnx); i++)
    grpindex[i] = i;

  /* Check whether we have to GCT stuff */
  bTCR = ftp2bSet(efGCT,nfile,fnm);
  if (MASTER(cr) && bTCR)
    fprintf(stderr,"Will do General Coupling Theory!\n");

  /* Initialize pull code */
  init_pull(log,nfile,fnm,&pulldata,state->x,mdatoms,inputrec->opts.nFreeze,
	    state->box,START(nsb),HOMENR(nsb),cr);
  
  if (repl_ex_nst > 0)
    repl_ex = init_replica_exchange(log,mcr,state,inputrec,
				    repl_ex_nst,repl_ex_seed);
  
  if (!inputrec->bUncStart && !bRerunMD) 
    do_shakefirst(log,ener,inputrec,nsb,mdatoms,state,vold,buf,f,
		  graph,cr,&mynrnb,grps,fr,top,edyn,&pulldata);
  debug_gmx();
    
  /* Compute initial EKin for all.. */
  if (grps->cosacc.cos_accel == 0)
    calc_ke_part(START(nsb),HOMENR(nsb),state->v,&(inputrec->opts),
		 mdatoms,grps,&mynrnb,state->lambda);
  else
    calc_ke_part_visc(START(nsb),HOMENR(nsb),
		      state->box,state->x,state->v,&(inputrec->opts),
		      mdatoms,grps,&mynrnb,state->lambda);
  debug_gmx();
	
  if (PAR(cr)) 
    global_stat(log,cr,ener,force_vir,shake_vir,
		&(inputrec->opts),grps,&mynrnb,nrnb,vcm,&terminate);
  debug_gmx();
  
  /* Calculate the initial half step temperature */
  temp0 = sum_ekin(TRUE,&(inputrec->opts),grps,ekin,NULL);

  debug_gmx();
  
  /* Initiate data for the special cases */
  if (bFFscan) {
    snew(xcopy,nsb->natoms);
    snew(vcopy,nsb->natoms);
    for(ii=0; (ii<nsb->natoms); ii++) {
      copy_rvec(state->x[ii],xcopy[ii]);
      copy_rvec(state->v[ii],vcopy[ii]);
    }
    copy_mat(state->box,boxcopy);
  }  
  
  /* Write start time and temperature */
  start_t=print_date_and_time(log,cr->nodeid,"Started mdrun");
  
  if (MASTER(cr)) {
    fprintf(log,"Initial temperature: %g K\n",temp0);
    if (bRerunMD) {
      fprintf(stderr,"starting md rerun '%s', reading coordinates from"
	      " input trajectory '%s'\n\n",
	      *(top->name),opt2fn("-rerun",nfile,fnm));
      if (bVerbose)
	fprintf(stderr,"Calculated time to finish depends on nsteps from "
		"run input file,\nwhich may not correspond to the time "
		"needed to process input trajectory.\n\n");
    } else
      fprintf(stderr,"starting mdrun '%s'\n%d steps, %8.1f ps.\n\n",
	      *(top->name),inputrec->nsteps,inputrec->nsteps*inputrec->delta_t);
  }

  /* Initialize values for invmass, etc. */
  update_mdatoms(mdatoms,state->lambda,TRUE);

  /* Set the node time counter to 0 after initialisation */
  start_time();
  debug_gmx();
  /***********************************************************
   *
   *             Loop over MD steps 
   *
   ************************************************************/
  
  /* if rerunMD then read coordinates and velocities from input trajectory */
  if (bRerunMD) {
    if (getenv("GMX_FORCE_UPDATE"))
      bForceUpdate = TRUE;

    bNotLastFrame = read_first_frame(&status,opt2fn("-rerun",nfile,fnm),
				     &rerun_fr,TRX_NEED_X | TRX_READ_V);
    if (rerun_fr.natoms != mdatoms->nr)
      gmx_fatal(FARGS,"Number of atoms in trajectory (%d) does not match the "
		  "run input file (%d)\n",rerun_fr.natoms,mdatoms->nr);
    if (inputrec->ePBC != epbcNONE) {
      if (!rerun_fr.bBox)
	gmx_fatal(FARGS,"Rerun trajectory frame step %d time %f does not contain a box, while pbc is used",rerun_fr.step,rerun_fr.time);
      if (max_cutoff2(rerun_fr.box) < sqr(fr->rlistlong))
	gmx_fatal(FARGS,"Rerun trajectory frame step %d time %f has too small box dimensions",rerun_fr.step,rerun_fr.time);
    }
  }

  /* loop over MD steps or if rerunMD to end of input trajectory */
  bFirstStep = TRUE;
  bLastStep = FALSE;
  bExchanged = FALSE;
  step = inputrec->init_step;
  step_rel = 0;
#ifdef USE_MPE
  MPE_Start_log( );
#endif
  while ((!bRerunMD && (step_rel <= inputrec->nsteps)) ||  
	 (bRerunMD && bNotLastFrame)) {
    
#ifdef USE_MPE
    /* MPI_Barrier( MPI_COMM_WORLD ); */
    MPE_Log_event( ev_timestep1, 0, "time step starts" );
#endif
    if (bRerunMD) {
      if (rerun_fr.bStep) {
	step = rerun_fr.step;
	step_rel = step - inputrec->init_step;
      }
      if (rerun_fr.bTime)
	t = rerun_fr.time;
      else
	t = step;
    } else {
      bLastStep = (step_rel == inputrec->nsteps);

      t = t0 + step*inputrec->delta_t;
    }
    
    do_log = do_per_step(step,inputrec->nstlog) || bLastStep;

    if (inputrec->efep != efepNO) {
      if (bRerunMD && rerun_fr.bLambda && (inputrec->delta_lambda!=0))
	state->lambda = rerun_fr.lambda;
      else
	state->lambda = lam0 + step*inputrec->delta_lambda;
    }
    if (MASTER(cr) && do_log && !bFFscan)
      print_ebin_header(log,step,t,state->lambda);

    if (bSimAnn) 
      update_annealing_target_temp(&(inputrec->opts),t);
    
    if (bRerunMD) {
      for(i=0; i<mdatoms->nr; i++)
	copy_rvec(rerun_fr.x[i],state->x[i]);
      if (rerun_fr.bV)
	for(i=0; i<mdatoms->nr; i++)
	  copy_rvec(rerun_fr.v[i],state->v[i]);
      else {
	for(i=0; i<mdatoms->nr; i++)
	    clear_rvec(state->v[i]);
	if (bRerunWarnNoV) {
	  fprintf(stderr,"\nWARNING: Some frames do not contain velocities.\n"
		  "         Ekin, temperature and pressure are incorrect,\n"
		  "         the virial will be incorrect when constraints are present.\n"
		  "\n");
	  bRerunWarnNoV = FALSE;
	}
      }
      copy_mat(rerun_fr.box,state->box);
      
      /* for rerun MD always do Neighbour Searching */
      bNS = ((inputrec->nstlist!=0) || bFirstStep);
    } else {
      /* Determine whether or not to do Neighbour Searching */
      bNS = ((inputrec->nstlist && (step % inputrec->nstlist==0 || bExchanged))
	     || bFirstStep);
    }
    
    /* Stop Center of Mass motion */
    bStopCM = do_per_step(step,abs(inputrec->nstcomm));

    /* Copy back starting coordinates in case we're doing a forcefield scan */
    if (bFFscan) {
      for(ii=0; (ii<nsb->natoms); ii++) {
	copy_rvec(xcopy[ii],state->x[ii]);
	copy_rvec(vcopy[ii],state->v[ii]);
      }
      copy_mat(boxcopy,state->box);
    }

    if (bVsites) {
      if (graph) {
	/* Following is necessary because the graph may get out of sync
	 * with the coordinates if we only have every N'th coordinate set
	 */
	if (bRerunMD || bExchanged)
	  mk_mshift(log,graph,state->box,state->x);
	shift_self(graph,state->box,state->x);
      }
      construct_vsites(log,state->x,&mynrnb,inputrec->delta_t,state->v,
		       &top->idef,graph,cr,fr->ePBC,state->box,vsitecomm);
      if (graph)
	unshift_self(graph,state->box,state->x);
    }
    debug_gmx();
    
    /* Set values for invmass etc. This routine not parallellized, but hardly
     * ever used, only when doing free energy calculations.
     */
    if(inputrec->efep != efepNO)
      update_mdatoms(mdatoms,state->lambda,FALSE); 
    
    clear_mat(force_vir);
    
    /* Ionize the atoms if necessary */
    if (bIonize)
      ionize(log,mdatoms,top->atoms.atomname,t,inputrec,state->x,state->v,
	     START(nsb),START(nsb)+HOMENR(nsb),state->box,cr);
      
    /* Update force field in ffscan program */
    if (bFFscan) {
      if (update_forcefield(nfile,fnm,fr,mdatoms->nr,state->x,state->box)) {
	if (gmx_parallel_env)
	  gmx_finalize(cr);
	exit(0);
      }
    }
    
    if (bShell_FlexCon) {
      /* Now is the time to relax the shells */
      count=relax_shells(log,cr,mcr,bVerbose,bFFscan ? step+1 : step,
			 inputrec,bNS,bStopCM,top,ener,fcd,
			 state,vold,vt,f,buf,mdatoms,nsb,&mynrnb,graph,
			 grps,
			 nshell,shells,nflexcon,fr,traj,t,mu_tot,
			 nsb->natoms,&bConverged,bVsites,vsitecomm,
			 fp_field);
      tcount+=count;
      
      if (bConverged)
	nconverged++;
    }
    else {
      /* The coordinates (x) are shifted (to get whole molecules) in do_force
       * This is parallellized as well, and does communication too. 
       * Check comments in sim_util.c
       */
		
      do_force(log,cr,mcr,inputrec,nsb,step,&mynrnb,top,grps,
	       state->box,state->x,f,buf,mdatoms,ener,fcd,bVerbose && !PAR(cr),
	       state->lambda,graph,
	       TRUE,bNS,FALSE,TRUE,fr,mu_tot,FALSE,t,fp_field,edyn);
    }
   
	
    if (bTCR)
      mu_aver=calc_mu_aver(cr,nsb,state->x,mdatoms->chargeA,mu_tot,top,mdatoms,
			   gnx,grpindex);
    
    if (bGlas)
      do_glas(log,START(nsb),HOMENR(nsb),state->x,f,
	      fr,mdatoms,top->idef.atnr,inputrec,ener);
    
    if (bTCR && bFirstStep) {
      tcr=init_coupling(log,nfile,fnm,cr,fr,mdatoms,&(top->idef));
      fprintf(log,"Done init_coupling\n"); 
      fflush(log);
    }
    
    /* Now we have the energies and forces corresponding to the 
     * coordinates at time t. We must output all of this before
     * the update.
     * for RerunMD t is read from input trajectory
     */
    if (bVsites) 
      spread_vsite_f(log,state->x,f,&mynrnb,&top->idef,
		     fr,graph,state->box,vsitecomm,cr);
      
    /* Calculation of the virial must be done after vsites!    */
    /* Question: Is it correct to do the PME forces after this? */
	/*>>> MFFA >>>*/
    if (!pdsb -> bPC)
		calc_virial(log,START(nsb),HOMENR(nsb),state->x,f,
				force_vir,fr->vir_el_recip,graph,state->box,&mynrnb,fr);
	/*<<< MFFA <<<*/
    /* Spread the LR force on virtual sites to the other particles... 
     * This is parallellized. MPI communication is performed
     * if the constructing atoms aren't local.
     */
    if (bVsites && fr->bEwald) 
      spread_vsite_f(log,state->x,fr->f_el_recip,&mynrnb,&top->idef,
		     fr,graph,state->box,vsitecomm,cr);
    
    sum_lrforces(f,fr,START(nsb),HOMENR(nsb));

    xx = (do_per_step(step,inputrec->nstxout) || bLastStep) ? state->x : NULL;
    vv = (do_per_step(step,inputrec->nstvout) || bLastStep) ? state->v : NULL;
    ff = (do_per_step(step,inputrec->nstfout)) ? f : NULL;

    fp_trn = write_traj(log,cr,traj,nsb,step,t,state->lambda,
			nrnb,nsb->natoms,xx,vv,ff,state->box);
    debug_gmx();
    
    /* don't write xtc and last structure for rerunMD */
    if (!bRerunMD && !bFFscan) {
      if (do_per_step(step,inputrec->nstxtcout))
	write_xtc_traj(log,cr,xtc_traj,nsb,mdatoms,
		       step,t,state->x,state->box,inputrec->xtcprec);
      if (bLastStep && MASTER(cr)) {
	fprintf(stderr,"Writing final coordinates.\n");
	write_sto_conf(ftp2fn(efSTO,nfile,fnm),
		       *top->name, &(top->atoms),state->x,state->v,state->box);
      }
      debug_gmx();
    }
    clear_mat(shake_vir);
    
    /* Afm and Umbrella type pulling happens before the update, 
     * other types in update 
     */
    if (pulldata.bPull && 
	(pulldata.runtype == eAfm || pulldata.runtype == eUmbrella))
      pull(&pulldata,state->x,f,force_vir,state->box,
	   top,inputrec->delta_t,step,t,
	   mdatoms,START(nsb),HOMENR(nsb),cr); 

	/*>>> MFFA >>>*/
	/* position of the boundary force after virial calculation and before 
	 * update. One loop extra but more clear */
	if (pdsb->bSDB) {
		/* include boundary force */
		do_sdb_force(START(nsb), HOMENR(nsb), f, state->x, pdsb, 
				state->box, force_vir);
		/* sum all force and pressure contributions over the nodes */
		if (PAR(cr))
			mpi_bound_force(pdsb->part_Fb, pdsb->Pb_sphere, pdsb->Pb_plane, 
					pdsb->PNUMB,cr);
	}
	/*<<< MFFA <<<*/

    if (bFFscan)
      clear_rvecs(nsb->natoms,buf);

    /* Box is changed in update() when we do pressure coupling,
     * but we should still use the old box for energy corrections and when
     * writing it to the energy file, so it matches the trajectory files for
     * the same timestep above. Make a copy in a separate array.
     */
    copy_mat(state->box,lastbox);
    
    /* This is also parallellized, but check code in update.c */
    /* bOK = update(nsb->natoms,START(nsb),HOMENR(nsb),step,state->lambda,&ener[F_DVDL], */
    bOK = TRUE;
    if (!bRerunMD || rerun_fr.bV || bForceUpdate) {
      dvdl = 0;
      update(nsb->natoms,START(nsb),HOMENR(nsb),step,&dvdl,
	     inputrec,mdatoms,state,graph,f,vold,
	     top,grps,shake_vir,cr,&mynrnb,edyn,&pulldata,bNEMD,
	     TRUE,bFirstStep,NULL,pres);
      if (fr->bSepDVDL && log && do_log)
	fprintf(log,sepdvdlformat,"Constraint",0.0,dvdl);
      ener[F_DVDL] += dvdl;
    } else {
      /* Need to unshift here */
      if ((inputrec->ePBC == epbcXYZ) && (graph->nnodes > 0))
	unshift_self(graph,state->box,state->x);
    }
    if (!bOK && !bFFscan)
      gmx_fatal(FARGS,"Constraint error: Shake, Lincs or Settle could not solve the constrains");

    /* Correct the new box if it is too skewed */
    if (DYNAMIC_BOX(*inputrec) && !bRerunMD)
      correct_box(state->box,fr,graph);
    /* (un)shifting should NOT be done after this,
     * since the box vectors might have changed
     */

    /* Non-equilibrium MD: this is parallellized, but only does communication
     * when there really is NEMD.
     */
    if (PAR(cr) && bNEMD) 
      accumulate_u(cr,&(inputrec->opts),grps);

	/*>>> MFFA >>>*/
	/*******************************************/
	/*************BOUNDARY PRESSURE*************/
	/*******************************************/
	/* note that continuation isn't exact */
	/* only for enclosed spherical systems, not for planes */
	
	/* volume calculation */
	rin = 0;
	rout = 0;
	volume = calc_volume(pdsb,lastbox,&rin,&rout);
	
	/* PC */
	if (pdsb -> bPC && ! bFirstStep) {
		/* F/A */
		copy_rvec(pdsb->SDB[0].cutoffs.center,center);
		pdsb->P_int = trace(pres) / DIM;
		/* coupling to outer virial pressure */
		if (inputrec->compress[ZZ][ZZ]!=0) {
			/* rout,rin correspond to PNUMB = 0 */
			/* Note really efficient can better use the Mu in update.c. If 
			 * I have time.*/ 
			/* implement for cavity growing within the vesicle */
			if (pdsb -> bQUENCH) {
				/* piston pressure coupling: no scaling */
				/* userint != 0, quench routine with both of the boundaries 
				 * scaling. This can be used for quenching big vesicles to 
				 * equ pressure */
				
				/* Now the cavity pressure is the virial pressure */
				if (inputrec->userint1 == 2)
					pdsb->Pb_sphere[0]=pdsb->P_int;
				sphere_piston(state->x, START(nsb), HOMENR(nsb), 
						pdsb->part_Fb[0].AREA_out, pdsb->Pb_sphere[1], 
						pdsb->part_Fb[0].AREA_in, pdsb->Pb_sphere[0], 
						inputrec->ref_p[XX][XX], &shift_out, &shift_in, 
						inputrec->delta_t, PARM, inputrec->opts.ref_t[0]);
				/* growth inner cavity with constant rate */
				if (inputrec->userint1 == 0) 
					shift_in = inputrec->userreal1 * inputrec->delta_t;
				
				/* set new rin en out */
				if (rin != 0)
					rin  += shift_in;
				if (rout != 0)
					rout += shift_out;
			} else
				/* mu scaling */
				sdb_pc(START(nsb), HOMENR(nsb), &rout, &rin, 
						inputrec->ref_p[XX][XX], pdsb->P_int, 
						inputrec->delta_t, inputrec->tau_p, &shift_out, 
						&shift_in, inputrec->compress[XX][XX], state->x, 
						center);
			
			/* carefull!!!! */
			/* n=0 gives the reference shell the rest is shifted along 
			 * with it*/
			for (n=0; n<pdsb->PNUMB; n++) {
				/* This approach is used to handle the shifting of 
				 * restraining shells within the system */
				/* scale the shell together with only the outher or inner 
				 * shell if the shell is to thin */
				if (pdsb->SDB[n].TYPE == efSPHERE) {
					correct_shell(&pdsb->SDB[0].cutoffs, 
							&pdsb->SDB[n].cutoffs, &shift_in,&shift_out);
					if (pdsb->SDB[n].cutoffs.pos_out != 0)
						pdsb->SDB[n].cutoffs.pos_out += shift_out ;
					if (pdsb->SDB[n].cutoffs.pos_in !=0)
						pdsb->SDB[n].cutoffs.pos_in += shift_in;
				}
			}
		}
		P_out_interval += pdsb->Pb_sphere[1];
		P_in_interval  += pdsb->Pb_sphere[0];
		P_out_tot += pdsb->Pb_sphere[1];
		P_in_tot  += pdsb->Pb_sphere[0];
		P_vir_interval += pdsb -> P_int;
		P_vir_tot += pdsb -> P_int;
		count1++;
		count2++;
		
		if (do_per_step(step,inputrec->nstorireout)) {
			if (MASTER(cr)) {
				fprintf(stderr,"time: %f (ps)\n",t);
				fprintf(stderr,"bound pres in: %f pres out: %f \n", 
						pdsb->Pb_sphere[0], pdsb->Pb_sphere[1]);
				fprintf(stderr,"bound pres  (interval) in : %f out: %f\n", 
						(float)(P_in_interval/count1), 
						(float)(P_out_interval/count1));
				fprintf(stderr,"pres gem (total) in: %f  out: %f\n", 
						(float)(P_in_tot/(count2)), 
						(float)(P_out_tot/(count2)));
				/* reset interval pressure */
				fprintf(stderr,"pres vir (interval) : %f", 
						(float)(P_vir_interval/count1));
				fprintf(stderr,"   pres vir (total) : %f\n", 
						(float)(P_vir_tot/count2));
				fprintf(stderr,"inner radius: %f outer: %f\n", rin, rout);
				fprintf(stderr,"Volume: %f (nm^3) \n", volume);
				fprintf(stderr,"Density: %f (kg/m^3)\n", 
						(float)(((1.66055*mdatoms->tmass)/volume)));
			}
			count1 = 0;
			P_in_interval = 0;
			P_out_interval = 0;
			P_vir_interval = 0;
		}
	}
	/*<<< MFFA <<<*/
 
    debug_gmx();
    if (grps->cosacc.cos_accel == 0)
      calc_ke_part(START(nsb),HOMENR(nsb),state->v,&(inputrec->opts),
		   mdatoms,grps,&mynrnb,state->lambda);
    else
      calc_ke_part_visc(START(nsb),HOMENR(nsb),
			state->box,state->x,state->v,&(inputrec->opts),
			mdatoms,grps,&mynrnb,state->lambda);

    /* since we use the new coordinates in calc_ke_part_visc, we should use
     * the new box too. Still, won't this be offset by one timestep in the
     * energy file? / EL 20040121
     */ 
    
    debug_gmx();
    /* Calculate center of mass velocity if necessary, also parallellized */
    if (bStopCM && !bFFscan && !bRerunMD)
      calc_vcm_grp(log,START(nsb),HOMENR(nsb),mdatoms->massT,
		   state->x,state->v,vcm);

    /* Check whether everything is still allright */    
    if (bGotTermSignal || bGotUsr1Signal) {
      if (bGotTermSignal)
	terminate = 1;
      else
	terminate = -1;
      fprintf(log,"\n\nReceived the %s signal\n\n",
	      bGotTermSignal ? "TERM" : "USR1");
      fflush(log);
      if (MASTER(cr)) {
	fprintf(stderr,"\n\nReceived the %s signal\n\n",
	      bGotTermSignal ? "TERM" : "USR1");
	fflush(stderr);
      }
      bGotTermSignal = FALSE;
      bGotUsr1Signal = FALSE;
    }

    if (PAR(cr)) {
      /* Globally (over all NODEs) sum energy, virial etc. 
       * This includes communication 
       */
      global_stat(log,cr,ener,force_vir,shake_vir,
		  &(inputrec->opts),grps,&mynrnb,nrnb,vcm,&terminate);
      /* Correct for double counting energies, should be moved to 
       * global_stat 
       */
      if (fr->bTwinRange && !bNS) 
	for(i=0; (i<grps->estat.nn); i++) {
	  grps->estat.ee[egCOULLR][i] /= cr->nnodes;
	  grps->estat.ee[egLJLR][i]   /= cr->nnodes;
	}
    }
    else
      cp_nrnb(&(nrnb[0]),&mynrnb);
    
    /* This is just for testing. Nothing is actually done to Ekin
     * since that would require extra communication.
     */
    if (!bNEMD && debug && (vcm->nr > 0))
      correct_ekin(debug,START(nsb),START(nsb)+HOMENR(nsb),state->v,
		   vcm->group_p[0],
		   mdatoms->massT,mdatoms->tmass,ekin);
    
    if ((terminate != 0) && (step - inputrec->init_step < inputrec->nsteps)) {
      if (terminate<0 && inputrec->nstxout)
	/* this is the USR1 signal and we are writing x to trr, 
	   stop at next x frame in trr */
	inputrec->nsteps =
	  (step/inputrec->nstxout + 1) * inputrec->nstxout - inputrec->init_step;
      else
	inputrec->nsteps = step + 1 - inputrec->init_step;
      fprintf(log,"\nSetting nsteps to %d\n\n",inputrec->nsteps);
      fflush(log);
      if (MASTER(cr)) {
	fprintf(stderr,"\nSetting nsteps to %d\n\n",inputrec->nsteps);
	fflush(stderr);
      }
      /* erase the terminate signal */
      terminate = 0;
    }
      
    /* Do center of mass motion removal */
    if (bStopCM && !bFFscan && !bRerunMD) {
      /* The last parameter is the temperature of the center of mass.
       * In this case we put 1 there, to assure that the temperature of
       * COM motion for each group should be below 1 K. Obviously
       * one could implement it as a tolerance times the actual T as well
       * e.g. 1e-3 T, but it remains arbitrary anyway. DvdS 2007-07-14.
       */ 
      check_cm_grp(log,vcm,1);
      do_stopcm_grp(log,START(nsb),HOMENR(nsb),state->x,state->v,vcm);
      inc_nrnb(&mynrnb,eNR_STOPCM,HOMENR(nsb));
      /*
      calc_vcm_grp(log,START(nsb),HOMENR(nsb),mdatoms->massT,x,v,vcm);
      check_cm_grp(log,vcm);
      do_stopcm_grp(log,START(nsb),HOMENR(nsb),x,v,vcm);
      check_cm_grp(log,vcm);
      */
    }
        
    /* Add force and shake contribution to the virial */
    m_add(force_vir,shake_vir,total_vir);

    /* Sum the potential energy terms from group contributions */
    sum_epot(&(inputrec->opts),grps,ener);

    /* Calculate the amplitude of the cosine velocity profile */
    grps->cosacc.vcos = grps->cosacc.mvcos/mdatoms->tmass;

    /* Sum the kinetic energies of the groups & calc temp */
    ener[F_TEMP]=sum_ekin(bRerunMD,&(inputrec->opts),grps,ekin,
			  &(ener[F_DVDLKIN]));
    ener[F_EKIN]=trace(ekin);
      
    /* Calculate Temperature coupling parameters lambda and adjust
     * target temp when doing simulated annealing
     */
    /*
    if(inputrec->etc==etcBERENDSEN)
      berendsen_tcoupl(&(inputrec->opts),grps,inputrec->delta_t);
    else if(inputrec->etc==etcNOSEHOOVER)
      nosehoover_tcoupl(&(inputrec->opts),grps,inputrec->delta_t);
    */

    /* Calculate pressure and apply LR correction if PPPM is used.
     * Use the box from last timestep since we already called update().
     */
	/*>>> MFFA >>>*/
	calc_pres(fr->ePBC,volume,ekin,total_vir,pres,
	      (fr->eeltype==eelPPPM) ? ener[F_COUL_RECIP] : 0.0);
	/*<<< MFFA <<<*/
    
    /* Calculate long range corrections to pressure and energy */
    if (bTCR || bFFscan)
      set_avcsixtwelve(log,fr,mdatoms,&top->atoms.excl);
    
    /* Calculate long range corrections to pressure and energy */
    calc_dispcorr(log,inputrec,fr,step,mdatoms->nr,lastbox,state->lambda,
		  pres,total_vir,ener);

    ener[F_ETOT]=ener[F_EPOT]+ener[F_EKIN];
    
    /* Check for excessively large energies */
    if (bIonize) {
#ifdef GMX_DOUBLE
      real etot_max = 1e200;
#else
      real etot_max = 1e30;
#endif
      if (fabs(ener[F_ETOT]) > etot_max) {
	fprintf(stderr,"Energy too large (%g), giving up\n",ener[F_ETOT]);
	break;
      }
    }

    /* The coordinates (x) were unshifted in update */
    if (bFFscan && (!bShell_FlexCon || bConverged))
      if (print_forcefield(log,ener,HOMENR(nsb),f,buf,xcopy,
			   &(top->blocks[ebMOLS]),mdatoms->massT,pres)) {
	if (gmx_parallel_env)
	  gmx_finalize(cr);
	fprintf(stderr,"\n");
	exit(0);
      }
    
    if (bTCR) {
      /* Only do GCT when the relaxation of shells (minimization) has converged,
       * otherwise we might be coupling to bogus energies. 
       * In parallel we must always do this, because the other sims might
       * update the FF.
       */
      
      /* Since this is called with the new coordinates state->x, I assume
       * we want the new box state->box too. / EL 20040121
       */
      do_coupling(log,nfile,fnm,tcr,t,step,ener,fr,
		  inputrec,MASTER(cr),
		  mdatoms,&(top->idef),mu_aver,
		  top->blocks[ebMOLS].nr,(mcr!=NULL) ? mcr : cr,
		  state->box,total_vir,pres,
		  mu_tot,state->x,f,bConverged);
      debug_gmx();
    }

    /* Time for performance */
    if (((step % stepout) == 0) || bLastStep)
      update_time();

    /* Output stuff */
    if (MASTER(cr)) {
      bool do_ene,do_dr,do_or,do_dihr;
      
      upd_mdebin(mdebin,fp_dgdl,mdatoms->tmass,step_rel,t,ener,state,lastbox,
		 shake_vir,force_vir,total_vir,pres,grps,mu_tot);
      do_ene = do_per_step(step,inputrec->nstenergy) || bLastStep;
      do_dr  = do_per_step(step,inputrec->nstdisreout) || bLastStep;
      do_or  = do_per_step(step,inputrec->nstorireout) || bLastStep;
      do_dihr= do_per_step(step,inputrec->nstdihreout) || bLastStep;
      print_ebin(fp_ene,do_ene,do_dr,do_or,do_dihr,do_log?log:NULL,
		 step,step_rel,t,
		 eprNORMAL,bCompact,mdebin,fcd,&(top->atoms),&(inputrec->opts));
      if (bVerbose)
	fflush(log);
	/*>>> MFFA >>>*/
	if ((do_per_step(step,inputrec->nstvout) || bLastStep) 
			&& (step != 0 && !bUPDSDB ))
		if (MASTER(cr) && (pdsb->bSDB)) {
			char **SDB_FILE_PARTS = NULL;
			split_string(opt2fn("-sb",nfile,fnm),&SDB_FILE_PARTS,&pdsb->PNUMB);
			for (i=0; i<pdsb->PNUMB; i++) {
				if (pdsb->SDB[i].TYPE == efSPHERE) {
					fprintf(stderr,"Dump: %s at time: %f\n ", 
							SDB_FILE_PARTS[i], t);
					dump_sdb_settings(pdsb->SDB[i], SDB_FILE_PARTS[i], 
							pdsb->bQUENCH, t);
				}
			}
		}
	}
	/*<<< MFFA <<<*/

    /* Remaining runtime */
    if (MASTER(cr) && bVerbose && ( ((step % stepout)==0) || bLastStep)) {
      if (bShell_FlexCon)
	fprintf(stderr,"\n");
      print_time(stderr,start_t,step,inputrec);
    }

    bExchanged = FALSE;
    if ((repl_ex_nst > 0) && (step > 0) && !bLastStep &&
	do_per_step(step,repl_ex_nst))
      bExchanged = replica_exchange(log,mcr,repl_ex,state,ener[F_EPOT],step,t);
    
    bFirstStep = FALSE;

    if (bRerunMD) 
      /* read next frame from input trajectory */
      bNotLastFrame = read_next_frame(status,&rerun_fr);

    if (!bRerunMD || !rerun_fr.bStep) {
      /* increase the MD step number */
      step++;
      step_rel++;
    }
  }
  /* End of main MD loop */
#ifdef USE_MPE
  MPE_Stop_log( );
#endif
  debug_gmx();

  /* Dump the NODE time to the log file on each node */
  if (PAR(cr)) {
    double *ct,ctmax,ctsum;
    
    snew(ct,cr->nnodes);
    ct[cr->nodeid] = node_time();
    gmx_sumd(cr->nnodes,ct,cr);
    ctmax = ct[0];
    ctsum = ct[0];
    for(i=1; (i<cr->nodeid); i++) {
      ctmax = max(ctmax,ct[i]);
      ctsum += ct[i];
    }
    ctsum /= cr->nnodes;
    fprintf(log,"\nTotal NODE time on node %d: %g\n",cr->nodeid,ct[cr->nodeid]);
    fprintf(log,"Average NODE time: %g\n",ctsum);
    fprintf(log,"Load imbalance reduced performance to %3d%% of max\n",
	    (int) (100.0*ctmax/ctsum));
    sfree(ct);
  }
      
  if (bRerunMD)
    close_trj(status);
	  
  if (MASTER(cr)) {
    print_ebin(fp_ene,FALSE,FALSE,FALSE,FALSE,log,step,step_rel,t,
	       eprAVER,FALSE,mdebin,fcd,&(top->atoms),&(inputrec->opts));
    print_ebin(fp_ene,FALSE,FALSE,FALSE,FALSE,log,step,step_rel,t,
	       eprRMS,FALSE,mdebin,fcd,&(top->atoms),&(inputrec->opts));
    close_enx(fp_ene);
    if (!bRerunMD && inputrec->nstxtcout)
      close_xtc_traj();
    close_trn(fp_trn);
    if (fp_dgdl)
      fclose(fp_dgdl);
    if (fp_field)
      fclose(fp_field);
  }
  debug_gmx();
  
  /* clean up edsam stuff, no effect if edyn->bEdsam == FALSE */
  finish_edsam(stdlog,top,inputrec,mdatoms,START(nsb),HOMENR(nsb),cr,edyn);

  if (bShell_FlexCon) {
    fprintf(log,"Fraction of iterations that converged:           %.2f %%\n",
	    (nconverged*100.0)/(step_rel+1));
    fprintf(log,"Average number of force evaluations per MD step: %.2f\n",
	    tcount/(step_rel+1));
  }

  if (repl_ex_nst > 0)
    print_replica_exchange_statistics(log,repl_ex);
  
  return start_t;
}
