/*
 * NAME
 *   g_bilayer_flux
 *     Calculate ion fluxes through a lipid bilayer.
 * SYNOPSIS
 *   g_bilayer_flux -s topol.tpr -f traj.xtc -o flux.xvg
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 * 
 * ---
 * Copyright (C) 2008  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>

#include "macros.h"
#include "smalloc.h"
#include "typedefs.h"
#include "names.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "strdb.h"
#include "vec.h"
#include "macros.h"
#include "index.h"
#include "pbc.h"
#include "xvgr.h"
#include "futil.h"
#include "matio.h"

typedef real rpair[2];

typedef struct {
	real upper, lower, far_up, far_down, border;
	bool wrap_up, wrap_down, wrap_far_up, wrap_far_down;
} LimitSet;

real membrane_position (int size_bilayer, atom_id index_bilayer[], rvec x[], 
		int axis, bool bProtein, int size_protein, atom_id index_protein[], 
		real nhood, bool verbose) {
	real avg=0.0, err=0.0;
	int  i, j, count=0;
	rvec r;

	if (bProtein) {
		/* calculate average bilayer position around the protein */
		for (i=0; i < size_bilayer; i++) {
			for (j=0; j < size_protein; j++) {
				rvec_sub(x[index_bilayer[i]], x[index_protein[j]], r);
				if ( norm(r) < nhood ) {
					avg += x[index_bilayer[i]][axis];
					count += 1;
					break;
				}
			}
		}
		avg /= count;
	} else {
		/* calculate average position of the bilayer */
		for (i=0; i < size_bilayer; i++) {
			avg += x[index_bilayer[i]][axis];
		}
		avg /= size_bilayer;
	}
	if (verbose)
		fprintf(stdout, "Average @ %g nm\n", avg);
	return avg;
}

LimitSet new_limit() {
	LimitSet limit;
	limit.upper = 0.0;
	limit.lower = 0.0;
	limit.far_up = 0.0;
	limit.far_down = 0.0;
	limit.border = 0.0;
	limit.wrap_up = FALSE;
	limit.wrap_down = FALSE;
	limit.wrap_far_up = FALSE;
	limit.wrap_far_down = FALSE;
	return limit;
}

LimitSet get_limits (real avg, real thickness, real cutoff, real ceiling, 
		bool verbose) {
	bool u=FALSE, d=FALSE;
	LimitSet limit;

	limit = new_limit();

	if (verbose)
		fprintf(stdout, "Ceiling @ %g nm\n", ceiling);
	if ((thickness + 2 * cutoff) > ceiling) 
		fprintf(stderr, "WARNING: not enough space in the box\n");
	/* membrane leaflets */
	limit.upper = avg + (thickness / 2.0);
	limit.lower = avg - (thickness / 2.0);
	if (limit.upper > ceiling) {
		limit.wrap_up = TRUE;
		limit.wrap_far_up = TRUE;
		limit.upper -= ceiling;
		u = TRUE;
	}
	if (limit.lower < 0.0) {
		limit.wrap_down = TRUE;
		limit.wrap_far_down = TRUE;
		limit.lower += ceiling;
		d = TRUE;
	}
	if (limit.wrap_up && limit.wrap_down) 
		fprintf(stderr, "WARNING: both upper and lower leaflet wrapped\n");
	if (verbose) {
		fprintf(stdout, "Upper leaflet @ %g nm\n", limit.upper);
		fprintf(stdout, "Lower leaflet @ %g nm\n", limit.lower);
	}
	/* upper and lower vicinity of the membrane */
	limit.far_up = limit.upper + cutoff;
	limit.far_down = limit.lower - cutoff;
	if (limit.far_up > ceiling) {
		if (u) 
			fprintf(stderr, "WARNING: wrapping twice around upper limit\n");
		limit.wrap_far_up = TRUE;
		limit.far_up -= ceiling;
	}
	if (limit.far_down < 0.0) {
		if (d)
			fprintf(stderr, "WARNING: wrapping twice around lower limit\n");
		limit.wrap_far_down = TRUE;
		limit.far_down += ceiling;
	}
	if (verbose) {
		fprintf(stdout, "Upper limit @ %g nm\n", limit.far_up);
		fprintf(stdout, "Lower limit @ %g nm\n", limit.far_down);
	}
	if (limit.wrap_up || limit.wrap_down || limit.wrap_far_up 
			|| limit.wrap_far_down) {
		limit.border = (limit.far_up + limit.far_down) / 2.0;
	} else {
		limit.border = (limit.far_up + limit.far_down) / 2.0;
		limit.border -= ceiling / 2.0;
		if (limit.border < 0.0) 
			limit.border += ceiling;
	}
	if (verbose) {
		fprintf(stdout, "Border @ %g nm\n", limit.border);
		if (limit.wrap_up) 
			fprintf(stderr, "Upper leaflet wrapped.\n");
		if (limit.wrap_down) 
			fprintf(stderr, "Lower leaflet wrapped.\n");
		if (limit.wrap_far_up) 
			fprintf(stderr, "Upper limit wrapped.\n");
		if (limit.wrap_far_down) 
			fprintf(stderr, "Lower limit wrapped.\n");
	}
	return limit;
}

int update_position (int *position, int size, atom_id index[], rvec x[], 
		int axis, LimitSet limit, bool verbose) {
	int i, channel=0;
	real curr;

	for (i=0; i < size; i++) {
		curr = x[index[i]][axis];
		if (limit.wrap_up || limit.wrap_down) {
			/* wrapped between leaflets */
			if (curr < limit.upper || curr > limit.lower) 
				position[i] = 0;      // within the leaflets
			else if (curr < limit.border) {
				if (curr < limit.far_up) 
					position[i] = 1;  // right above membrane
				else 
					position[i] = 2;  // far-away above membrane
			} else if (curr > limit.far_down) 
					position[i] = -1; // right under membrane
				else
					position[i] = -2; // far-away under membrane
		} else if (limit.wrap_far_up) {
			/* wrapped between bilayer and upper limit */
			if (curr > limit.lower && curr < limit.upper)
				position[i] = 0;
			else if (curr > limit.upper || curr < limit.far_up) 
				position[i] = 1;
			else if (curr < limit.border) 
				position[i] = 2;
			else if (curr > limit.far_down) 
				position[i] = -1;
			else
				position[i] = -2;
		} else if (limit.wrap_far_down) {
			/* wrapped between bilayer and lower limit */
			if (curr > limit.lower && curr < limit.upper)
				position[i] = 0;
			else if (curr < limit.lower || curr > limit.far_down) 
				position[i] = -1;
			else if (curr > limit.border) 
				position[i] = -2;
			else if (curr < limit.far_up) 
				position[i] = 1;
			else
				position[i] = 2;
		} else {
			/* no wrapping */
			if (curr > limit.lower && curr < limit.upper) 
				position[i] = 0;
			else if (curr > limit.upper) {
				if (curr < limit.far_up) 
					position[i] = 1;
				else 
					position[i] = 2;
			} else if (curr > limit.far_down) 
				position[i] = -1;
			else 
				position[i] = -2;
		}
		if (position[i] == 0) 
			channel++;
	}
	if (verbose)
		fprintf(stdout, "Atoms in channel: %d\n", channel);
	return channel;
}

void update_side (int *side, int *position, int size) {
	int i;

	for (i=0; i < size; i++) {
		if (position[i] == 1) 
			side[i] = 1;
		else if (position[i] == -1) 
			side[i] = -1;
		else if (position[i] != 0)
			side[i] = 0;
	}
}

int check_transitions (int *ups, int *downs, int *side, int *position, 
		int size, bool verbose) {
	int i, count=0;

	for (i=0; i < size; i++) {
		if (position[i] == 1 && side[i] == -1) {
			ups[i] += 1;
			count++;
		} else if (position[i] == -1 && side[i] == 1) {
			downs[i] += 1;
			count++;
		}
	}
	if (verbose)
		fprintf(stdout, "Transitions: %d\n", count);
	return count;
}

int gmx_bilayer_flux (int argc, char *argv[])
{
	static char *desc[] = {
		"Calculate ion fluxes through lipid bilayers with or without ",
		"channel proteins.[PAR]",
		"For each frame calculates the average center of bilayer, divides ",
		"the atoms into two groups on either side of this boundary, and ",
		"checks the number of ions passing through the boundary. The ", 
		"decision boundary is calculated on a NxN grid normal to the ", 
		"channel vector ([TT]-direction[tt]) as a mass weighted average ",
		"over all atoms selected interactively by the user. The ion group ",
		"is also selected by the user. The resolution of the grid ", 
		"is set by [TT]-resolution[tt]. The flux is determined by the ", 
		"number of transition events between current and previous frame."
	};
	bool        bFirst=TRUE;
	bool        bProtein=FALSE, verbose=FALSE;
	bool        bX=FALSE, bY=FALSE, bZ=TRUE;
	t_topology  top;
	t_atoms    *atoms;
	matrix      box;
	char        buf[255];
	rvec       *x;
	int         ePBC, status, natom, axis;
	int         size_bilayer, size_ions, size_protein;
	int         i, count, df, f=0;
	atom_id    *index_bilayer, *index_ions, *index_protein;
	char       *grpname_bilayer, *grpname_ions, *grpname_protein;
	real        t, t0, r, nhood=1.4, cutoff=3.0, thickness=2.0;
	real        middle;
	LimitSet    limits;
	int        *side, *position, *location, *ups, *downs;
	FILE       *fp=NULL;
	
	/* for index file */
	char     **gnames;
	t_block  *block;
	atom_id  *index_trans;

	char *keys[] = {"Flux", "sum(Flux)", "Channel"};

	t_pargs pa[] = {
		{ "-x", TRUE, etBOOL, {&bX}, 
			"membrane normal is along the x-axis" },
		{ "-y", TRUE, etBOOL, {&bY}, 
			"membrane normal is along the y-axis" },
		{ "-z", TRUE, etBOOL, {&bZ}, 
			"membrane normal is along the z-axis" },
		{ "-protein", TRUE, etBOOL, {&bProtein}, 
			"consider only lipids close to a memrane protein" },
		{ "-nhood", FALSE, etREAL, {&nhood},
			"consider atoms closer than nhood (nm) to be in the neighbourhood of the protein" },
		{ "-cutoff", FALSE, etREAL, {&cutoff},
			"exclude from flux calculations atoms beyond cutoff (nm) from the membrane" },
		{ "-thickness", FALSE, etREAL, {&thickness}, 
			"thickness of the membrane (nm), i.e. height of the membrane-channel" },
		{ "-v", FALSE, etBOOL, {&verbose}, 
			"be loud and noisy"}
	};
	t_filenm fnm[] = {
		{ efTRX, "-f", NULL, ffREAD },
		{ efTPS, "-s", NULL, ffREAD },
		{ efNDX, "-n", NULL, ffOPTRD },
		{ efXVG, NULL, "flux", ffWRITE },
		{ efNDX, "-no", "flux", ffOPTWR },
	};
#define NFILE asize(fnm)
	if (bX)
		axis = XX;
	else if (bY)
		axis = YY;
	else
		axis = ZZ;

	CopyRight(stderr, argv[0]);
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE, 
			NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);
	
	/* read topology */
	read_tps_conf(ftp2fn(efTPS, NFILE, fnm), buf, &top, &x, NULL, 
			box, FALSE);
	atoms = &(top.atoms);
	/* read bilayer and ion groups */
	fprintf(stderr, "Select group for the bilayer\n");
	get_index(atoms, opt2fn_null("-n", NFILE, fnm), 1, &size_bilayer, 
			&index_bilayer, &grpname_bilayer);
	fprintf(stderr, "Select group for the ions to monitor\n");
	get_index(atoms, opt2fn_null("-n", NFILE, fnm), 1, &size_ions, 
			&index_ions, &grpname_ions);
	/* read protein group? */
	if (bProtein) {
		fprintf(stderr, "Select group for the protein\n");
		get_index(atoms, opt2fn_null("-n", NFILE, fnm), 1, &size_protein, 
				&index_protein, &grpname_protein);
	}
	
	/* write header */
	fp = xvgropen(ftp2fn(efXVG, NFILE, fnm), 
			"Ion flux", "Time (ps)", "#");
	xvgr_legend(fp, 3, keys);
	
	natom = read_first_x(&status, ftp2fn(efTRX, NFILE, fnm), &t, &x, box);
	t0 = t;
	/* allocate memory */
	side = malloc(size_ions * sizeof(int) + 1);
	position = malloc(size_ions * sizeof(int) + 1);
	location = malloc(size_ions * sizeof(int) + 1);
	ups = malloc(size_ions * sizeof(int) + 1);
	downs = malloc(size_ions * sizeof(int) + 1);
	/* initialise, just to be safe */
	for (i=0; i < size_ions; i++) {
		side[i] = 0;
		position[i] = 0;
		location[i] = 0;
		ups[i] = 0;
		downs[i] = 0;
	}

	/* main loop */
	do {
		if (verbose) {
			fprintf(stdout, "\nTime: %g ps\n", t);
			fprintf(stdout, "Box: %g %g %g\n", 
					box[XX][XX], box[YY][YY], box[ZZ][ZZ]);
		}
		put_atoms_in_box(box, natom, x);
		if (bFirst) {
			middle = membrane_position(size_bilayer, index_bilayer, x, axis, 
					bProtein, size_protein, index_protein, nhood, verbose);
			limits = get_limits(middle, thickness, cutoff, box[axis][axis], 
					verbose);
			count = update_position(position, size_ions, index_ions, x, axis, 
					limits, verbose); 
			update_side(side, position, size_ions);
			bFirst = FALSE;
			df = 0;
		} else {
			middle = membrane_position(size_bilayer, index_bilayer, x, axis, 
					bProtein, size_protein, index_protein, nhood, verbose);
			limits = get_limits(middle, thickness, cutoff, box[axis][axis], 
					verbose);
			count = update_position(position, size_ions, index_ions, x, axis, 
					limits, verbose); 
			df = check_transitions(ups, downs, position, side, size_ions, 
					verbose);
			update_side(side, position, size_ions);
		}
		f += df;
		fprintf(fp,"%g  %d  %d %d\n", t, df, f, count);
	} while (read_next_x(status, &t, natom, x, box));

	/* write an index file of fluxed ions  */
	snew(gnames,1);
	block = new_block();
    analyse(atoms, block, &gnames, FALSE, FALSE);
	
	snew(index_trans, size_ions);
	count=0;
	for (i=0; i < size_ions; i++) 
		if (ups[i] > 0)
			index_trans[count++] = index_ions[i];
	add_grp(block, &gnames, count, index_trans, "Fluxed+");
	for (i=0; i < size_ions; i++)
		index_trans[i] = 0;
	count=0;
	for (i=0; i < size_ions; i++) 
		if (downs[i] > 0)
			index_trans[count++] = index_ions[i];
	add_grp(block, &gnames, count, index_trans, "Fluxed-");
	
	write_index(opt2fn("-no", NFILE, fnm), block, gnames);

	/* report total flux */
	fprintf(stderr, "\n");
	fprintf(stdout, "Total flux: %d in %g ps\n", f, (t - t0));
	fprintf(stdout, "Flux rate: %g / ns\n", f / (t -t0) * 1000.0);
	
	/* free memory */
	free(side);
	free(position);
	free(location);
	free(ups);
	free(downs);

	/* close down */
	fclose(fp);
	close_trj(status);
	thanx(stderr);
	return 0;
}

