/*
 * NAME
 *   g_vesicle_flux
 *     Calculate ion fluxes through a lipid vesicle.
 * SYNOPSIS
 *   g_vesicle_flux -s topol.tpr -f traj.xtc -o flux.xvg
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 * 
 * ---
 * Copyright (C) 2008  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>

#include "macros.h"
#include "smalloc.h"
#include "typedefs.h"
#include "names.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "strdb.h"
#include "vec.h"
#include "macros.h"
#include "index.h"
#include "pbc.h"
#include "xvgr.h"
#include "futil.h"
#include "matio.h"


void calculate_com (rvec *com, int n, atom_id index[], rvec x[])
{
	int i;
	
	/* calculate the center of mass */
	/*   TODO: actually use masses! */
	(*com)[XX] = 0.0;
	(*com)[YY] = 0.0;
	(*com)[ZZ] = 0.0;
	for (i=0; i < n; i++) {
		rvec_add(*com, x[index[i]], *com);
	}
	svmul(1.0 / n, *com, *com);
}

real calculate_avg_radius (int n, atom_id index[], rvec x[], rvec *com)
{
	rvec v;
	real r=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], *com, v);
		r = r + norm(v);
	}
	r = r / n;
	return r;
}

int calculate_flux (int *in, int *out, int n, atom_id index[], rvec *com, 
		real r, rvec x[], bool bInverse)
{
	int i, prev, flux;
	real s;
	rvec v;

	/* previous number of ions inside */
	prev = *in;
	
	/* count ions */
	*in = 0;
	*out = 0;
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], *com, v);
		s = norm(v);
		if (s > r) 
			(*out)++;
		else 
			(*in)++;
	}

	/* invert flux direction? */
	if (bInverse)
		flux = (*in) - prev;
	else
		flux = prev - (*in);

	return flux;
}


int gmx_vesicle_flux (int argc, char *argv[])
{
	static char *desc[] = {
		"Calculate ion fluxes through lipid vesicles with or without ",
		"channel proteins.[PAR]",
		"For each frame calculates the average radius of the vesicle ",
		"and checks the number of ions inside and outside of this sphere. ",
		"The flux is then determined as the increase/decrease of inside ", 
		"ions from the previous time step. Positive flux direction is by ",
		"default from inside to outside, but this can be changed by ", 
		"[TT]-inverse[tt]. The average radius of the vesicle is ", 
		"calculated as a mass weighted average over all atoms selected ", 
		"interactively by the user. The ion group is also selected by the ",
		"user."
	};
	static bool bInverse=FALSE;
	bool        bFirst=TRUE;
	t_topology  top;
	t_atoms    *atoms;
	matrix      box;
	char        buf[255];
	rvec       *x, com;
	int         ePBC, status, natom, in=0, out=0, first=0;
	int         size_vesicle, size_ions;
	int         flux, fluxed=0;
	atom_id    *index_vesicle, *index_ions;
	char       *grpname_vesicle, *grpname_ions;
	real        t, r;
	FILE       *fp;
	
	char *keys[] = {"Flux", "sum(Flux)", "Inside", "Outside", "Radius"};
	
	t_pargs pa[] = {
		{ "-inverse", TRUE, etBOOL, {&bInverse}, 
			"inverse flux direction to be from outside to inside" }
	};
	t_filenm fnm[] = {
		{ efTRX, "-f", NULL, ffREAD },
		{ efTPS, "-s", NULL, ffREAD },
		{ efNDX, NULL, NULL, ffOPTRD },
		{ efXVG, NULL, "flux", ffWRITE },
	};
#define NFILE asize(fnm)
	CopyRight(stderr, argv[0]);
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE, 
			NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);
	/*parse_common_args(&argc, argv, 
			PCA_CAN_TIME | PCA_TIME_UNIT | PCA_CAN_VIEW | PCA_BE_NICE,
		    NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);*/
	
	/* read topology */
	read_tps_conf(ftp2fn(efTPS, NFILE, fnm), buf, &top, &x, NULL, 
			box, FALSE);
	atoms = &(top.atoms);
	/* read vesicle and ion groups */
	printf("Select group for the vesicle\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_vesicle, 
			&index_vesicle, &grpname_vesicle);
	printf("Select group for the ions to monitor\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_ions, 
			&index_ions, &grpname_ions);
	
	/* write header */
	fp = xvgropen(ftp2fn(efXVG, NFILE, fnm), 
			"Ion flux", "Time (ps)", "#");
	xvgr_legend(fp, 4, keys);

	/* main loop */
	natom = read_first_x(&status, ftp2fn(efTRX, NFILE, fnm), &t, &x, box);
	do {
		calculate_com(&com, size_vesicle, index_vesicle, x);
		r = calculate_avg_radius(size_vesicle, index_vesicle, x, com);
		flux = calculate_flux(&in, &out, size_ions, index_ions, com, r, x, 
				bInverse);
		if (bFirst) {
			flux = 0;
			bFirst = FALSE;
			first = in;
		}
		fluxed += flux;
		fprintf(fp,"%g  %d  %d  %d  %d  %g\n", t, flux, fluxed, in, out, r);
	} while (read_next_x(status, &t, natom, x, box));

	fprintf(stderr, "\n");
	fprintf(stdout, "Total flux: %d in %g ps\n", fluxed, t);
	fprintf(stdout, "Flux rate: %g / ns\n", fluxed / t * 1000.0);
	
	/* close down */
	fclose(fp);
	close_trj(status);
	thanx(stderr);
	return 0;
}

